#!/usr/bin/env python
# encoding: utf-8
"""
befriender_client.py

This is the main executable for the befriender client.

Created by Elliot Pennington on 2008-11-30. 
"""

from socket import *
from message import *
from Queue import *
from Tkinter import *
import tkMessageBox
import threading
import receiver, sender, gui
import sys
import net_info

host = "localhost"
port = 1465
buf_sz = 1024
addr = (host, port)

def connect_to_server():
    """ get a socket connection """
    sockfd = socket(AF_INET, SOCK_STREAM)
    sockfd.connect((net_info.addr, net_info.port))
    return sockfd

class BefrienderClient:
    
    """ 
    b_client represents an internal befriender cleint. A new instantiation
    of one of these is like a whole separate login. The flow of this class 
    will login based on a required username and password to the server, and
    then receive messages from the server. The b_client will listen and dispatch
    the messages to the appropriate window based on incoming screen name, or 
    spawn a new conversation thread to handle a new screen name. 
    
    Attributes:
        sockfd: The socket descriptor for the connection to the server
        username: The username for the login of this client
        current_convos: List of screen names (strings) with which this user is currently in 
                    conversation
        recv_queues: dictionary association of Queues, each associated with a conversation.
                    The dictionary key is the screen name of the conversation partner.  
        recv_events: dictionary association of Events, each associated with a conversation.
                    The keys are the same as for the queues.
        receiver: The method to run for the receiver thread. In this case, the graphical window
                    thread for printing conversations and running a window.
        send_queue: The Queue for outgoing messages and commands. Popped only within the sender 
                    thread.
        send_event: The Event for the send queue and the sender thread. 
        sender: The thread created solely for the purpose of sending messages and commands
                    to the server.
    """
    
    def __init__(self):
        """ 
        This constructor will serve forever. Instantiating an object of this type
        is all you should ever need to do with it. 
        """
        global addr
        global buf_sz
        
        self.sockfd = connect_to_server()
  
        self.current_convos = []
        self.recv_events = {}
        self.recv_queues = {}
        self.receiver = gui.TabbedFrame
        self.outer_event = threading.Event()
        
        self.send_queue = Queue()
        self.send_event = threading.Event()
        self.sender = threading.Thread(target=sender.Sender,
                                       args=(self.sockfd, self.send_queue, self.send_event))
        self.sender.start()
        
        self.username = ""
#Should probably thread this and/or wait on it to finish, right now it's a bit of a hack since the next window simply won't start till this is done on windows
        self.loginwindow = gui.login_window
        gui.login_window(self.send_queue, self.send_event, self.set_username, self.set_logout)
#To test login without window
#        self.send_queue.put("~server:login:%s:%s" % ("username", "password"), True)
#        self.send_event.set()
#        self.send_event.clear()

#Quick setup for the tab window
        self.root = Tk()
        
        self.root.protocol("WM_DELETE_WINDOW", self.close_UI)
        self.root.title("Befriender Client")
        
        self.win = gui.Window(self.root)
        self.menu = gui.menu_window(self.win(), self.username, self.send_queue, self.outer_event, self.set_logout)
        self.win.add_screen(self.menu, "Menu")
        
        self.options = gui.options_window(self.win(), self.username, self.send_queue, self.outer_event)
        self.win.add_screen(self.options, "Options")
               
        self.spawn_root_thread()
        
        self.wait_forever()
        
    def wait_forever(self):
        """ 
        Wait for messages as they come in from the server.
        
        If a message is from a new sender, make a new thread to handle
        that conversation. The new thread gets a new queue, created for IPC. 
        The new screen name is added to current_convos.
        """
        while (1):
            print "Waiting forever"

# A simple test to ensure win.add_screen works past the mainloop starting and with an entirely new frame
#    worked like a charm, so we should be good.
#            import time
#            time.sleep(5)          
#            self.options2 = gui.options_window(self.win(), "lol", self.send_queue, self.outer_event)
#            self.win.add_screen(self.options2, "test2")           
            data = self.sockfd.recv(buf_sz)
            msg = Message(data)
            my_sn = msg.getSn()
            my_msg = msg.getMsg()
            print ("Received Message: %s" % my_msg)
            print ("From SN: %s " % my_sn)
            if my_sn not in self.current_convos:
                self.recv_queues[my_sn] = Queue()
                self.recv_events[my_sn] = threading.Event()
                self.current_convos.append(my_sn)
                self.spawn_thread(self.win(), my_sn, 
                                  self.recv_queues[my_sn], 
                                  self.send_queue, 
                                  self.recv_events[my_sn], 
                                  self.send_event)
                
                #make sure the gui is about done
                self.outer_event.wait()
                #snap control back to the gui for the mainloop to start
                self.outer_event.set()
                self.outer_event.clear()
                import time
                time.sleep(1)
                self.recv_queues[my_sn].put(my_msg, True)
                
                #now we know we can set this and make the receiver write
                self.recv_events[my_sn].set()
                self.recv_queues[my_sn].join()
                self.recv_events[my_sn].clear()
            else:
                self.recv_queues[my_sn].put(my_msg, True)
                self.recv_events[my_sn].set()
#                self.recv_queues[my_sn].join()
                self.recv_events[my_sn].clear()

    def wait_foreverold(self):
        """ 
        Wait for messages as they come in from the server.
        
        If a message has is from a new sender, make a new thread to handle
        that conversation. The new thread gets a new queue, created for IPC. 
        The new screen name is added to current_convos. 
        """
        while (1):
            print "Waiting forever"
            data = self.sockfd.recv(buf_sz)
            msg = Message(data)
            my_sn = msg.getSn()
            my_msg = msg.getMsg()
            print ("Received Message: %s" % my_msg)
            print ("From SN: %s " % my_sn)
            if my_sn not in self.current_convos:
                self.recv_queues[my_sn] = Queue()
                self.recv_events[my_sn] = threading.Event()
                self.current_convos.append(my_sn)
                
                self.spawn_thread(self.win, my_sn, 
                                  self.recv_queues[my_sn], 
                                  self.send_queue, 
                                  self.recv_events[my_sn], 
                                  self.send_event)
                
                #make sure the gui is about done
                self.outer_event.wait()
                #snap control back to the gui for the mainloop to start
                self.outer_event.set()
                self.outer_event.clear()
                import time
                time.sleep(1)
                self.recv_queues[my_sn].put(my_msg, True)
                
                #now we know we can set this and make the receiver write
                self.recv_events[my_sn].set()
                self.recv_queues[my_sn].join()
                self.recv_events[my_sn].clear()
            else:
                self.recv_queues[my_sn].put(my_msg, True)
                self.recv_events[my_sn].set()
#                self.recv_queues[my_sn].join()
                self.recv_events[my_sn].clear()
                
    def spawn_thread(self, master, sn, in_queue, out_queue, in_event, out_event):
        """ Create a new thread for the new conversation """
                
        t = threading.Thread(target=self.process_request_thread,
                             args=(master, sn, in_queue, out_queue, in_event, out_event, self.outer_event))
        t.start()
    
    def process_request_thread(self, master, sn, in_queue, out_queue, in_event, out_event, top_event):
        """ Called from within the new thread. Runs the handler. """ 
        self.receiver(master, sn, in_queue, out_queue, in_event, out_event, top_event)
#I'm not sure if this is right
        self.win.add_screen(self.receiver, sn)        
        
    def spawn_root_thread(self):
        th = threading.Thread(target=self.process_root_thread,args=())
        th.start()
    
    def process_root_thread(self):
        self.root.mainloop()
    
    def set_username(self, username2):
        print "Username changed to: %s" % username2
        self.username = username2
    
    def set_logout(self):
        print "Logging out"
        self.root.destroy()
#TODO: close socket, windows more cleanly, etc.        
        sys.exit(0) 
    
    def close_UI(self):
        if tkMessageBox.askokcancel("Quit", "Are you sure you want to log out?"):
            self.send_queue.put("~server:logout", True)
            self.send_event.set()
            self.send_event.clear()
            self.set_logout()
        
if __name__ == '__main__':
    client = BefrienderClient()
#   client.login("username", "password")
#   client.process_incoming_messages()
   #test 
#    data = "Screenname:F:F:T: UU,:F:A:en:<HTML>Hello</HTML>"
#    msg = deserialize(data)
#    print "Received: "
#    msg.display()
#    client.receive_stuff(msg)

"""
TODO(esp):
Here's what I keep getting from the server, aside from piles of random unimportant TOC errors:
Waiting for connection!
DrizztDoUrden50:F:F:T:A ,:F:A:en:<HTML><BODY><FONT FACE="Arial" SIZE=2 COLOR=#000000>test</FONT></BODY></HTML>
Unhandled drizztdourden50: test
exception in thread started by <bound method BotManager._dispatcher of <toc.BotManager instance at 0x00BF8698>>
"""


"""
Basic server communication api:
~server:login:usr:pass
~server:history:user/screenname/both
~server:notes:add/retrieve:user:sn
~servper:settings:set:key:value
~server:settings:get:key
~server:logout
~server:available:[true|false]
~server:endConvo:sn
"""   